/*
 * Copyright 2004 Digi International (www.digi.com)
 *      Scott H Kilau <Scott_Kilau at digi dot com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * 
 *      NOTE TO LINUX KERNEL HACKERS:  DO NOT REFORMAT THIS CODE!
 *
 *      This is shared code between Digi's CVS archive and the
 *      Linux Kernel sources.
 *      Changing the source just for reformatting needlessly breaks
 *      our CVS diff history.
 *
 *      Send any bug fixes/changes to:  Eng.Linux at digi dot com.
 *      Thank you.
 *
 *
 * 
 * $Id: dgap_sysfs.c,v 1.3 2006/01/06 17:29:23 scottk Exp $   
 */


#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/serial_reg.h>
#include <linux/device.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
  
#include "dgap_driver.h"
#include "dgap_proc.h"
#include "dgap_mgmt.h"

/*
 * Unfortunately, I have no choice but to wrap this whole file around 
 * one big ifdef.
 *
 * Because of fundamental changes in the kernel SYSFS implementation
 * which ended up breaking all the existing SYSFS code in the driver,
 * I have no choice but to only fully support SYSFS on 2.6.13+...
 */
#ifdef DGAP_SYSFS_SUPPORT



static struct class *dgap_tty_class;



static ssize_t dgap_driver_version_show(struct device_driver *ddp, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", DG_PART);
}
static DRIVER_ATTR(version, S_IRUSR, dgap_driver_version_show, NULL);


static ssize_t dgap_driver_boards_show(struct device_driver *ddp, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%d\n", dgap_NumBoards);
}
static DRIVER_ATTR(boards, S_IRUSR, dgap_driver_boards_show, NULL);


static ssize_t dgap_driver_maxboards_show(struct device_driver *ddp, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%d\n", MAXBOARDS);
}
static DRIVER_ATTR(maxboards, S_IRUSR, dgap_driver_maxboards_show, NULL);


static ssize_t dgap_driver_pollcounter_show(struct device_driver *ddp, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%ld\n", dgap_poll_counter);
}
static DRIVER_ATTR(pollcounter, S_IRUSR, dgap_driver_pollcounter_show, NULL);


static ssize_t dgap_driver_state_show(struct device_driver *ddp, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", dgap_driver_state_text[dgap_driver_state]);
}
static DRIVER_ATTR(state, S_IRUSR, dgap_driver_state_show, NULL);


static ssize_t dgap_driver_debug_show(struct device_driver *ddp, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "0x%x\n", dgap_debug);
}

static ssize_t dgap_driver_debug_store(struct device_driver *ddp, const char *buf, size_t count)
{
	sscanf(buf, "0x%x\n", &dgap_debug);
	return count;
}
static DRIVER_ATTR(debug, (S_IRUSR | S_IWUSR), dgap_driver_debug_show, dgap_driver_debug_store);


static ssize_t dgap_driver_rawreadok_show(struct device_driver *ddp, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "0x%x\n", dgap_rawreadok);
}

static ssize_t dgap_driver_rawreadok_store(struct device_driver *ddp, const char *buf, size_t count)
{
	sscanf(buf, "0x%x\n", &dgap_rawreadok);
	return count;
}
static DRIVER_ATTR(rawreadok, (S_IRUSR | S_IWUSR), dgap_driver_rawreadok_show, dgap_driver_rawreadok_store);


static ssize_t dgap_driver_pollrate_show(struct device_driver *ddp, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%dms\n", dgap_poll_tick);
}

static ssize_t dgap_driver_pollrate_store(struct device_driver *ddp, const char *buf, size_t count)
{
	sscanf(buf, "%d\n", &dgap_poll_tick);
	return count;
}
static DRIVER_ATTR(pollrate, (S_IRUSR | S_IWUSR), dgap_driver_pollrate_show, dgap_driver_pollrate_store);


void dgap_create_driver_sysfiles(struct pci_driver *dgap_driver)
{
	struct device_driver *driverfs = &dgap_driver->driver;
	driver_create_file(driverfs, &driver_attr_version);
	driver_create_file(driverfs, &driver_attr_boards);
	driver_create_file(driverfs, &driver_attr_maxboards);
	driver_create_file(driverfs, &driver_attr_debug);
	driver_create_file(driverfs, &driver_attr_rawreadok); 
	driver_create_file(driverfs, &driver_attr_pollrate);
	driver_create_file(driverfs, &driver_attr_pollcounter);
	driver_create_file(driverfs, &driver_attr_state);
}


void dgap_remove_driver_sysfiles(struct pci_driver *dgap_driver)
{
	struct device_driver *driverfs = &dgap_driver->driver;
	driver_remove_file(driverfs, &driver_attr_version);
	driver_remove_file(driverfs, &driver_attr_boards);
	driver_remove_file(driverfs, &driver_attr_maxboards);
	driver_remove_file(driverfs, &driver_attr_debug);
	driver_remove_file(driverfs, &driver_attr_rawreadok);
	driver_remove_file(driverfs, &driver_attr_pollrate);
	driver_remove_file(driverfs, &driver_attr_pollcounter);
	driver_remove_file(driverfs, &driver_attr_state);
}


#define DGAP_VERIFY_BOARD(p, bd)			\
	if (!p)						\
		return (0);				\
							\
	bd = dev_get_drvdata(p);			\
	if (!bd || bd->magic != DGAP_BOARD_MAGIC)	\
		return (0);				\
	if (bd->state != BOARD_READY)			\
		return (0);				\


static ssize_t dgap_ports_state_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		count += snprintf(buf + count, PAGE_SIZE - count,
			"%d %s\n", bd->channels[i]->ch_portnum,
			bd->channels[i]->ch_open_count ? "Open" : "Closed");
	}
	return count;
}
static DEVICE_ATTR(ports_state, S_IRUSR, dgap_ports_state_show, NULL);


static ssize_t dgap_ports_baud_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		count +=  snprintf(buf + count, PAGE_SIZE - count,
			"%d %d\n", bd->channels[i]->ch_portnum, bd->channels[i]->ch_baud_info);
	}
	return count;
}
static DEVICE_ATTR(ports_baud, S_IRUSR, dgap_ports_baud_show, NULL);


static ssize_t dgap_ports_msignals_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		if (bd->channels[i]->ch_open_count) {
			count += snprintf(buf + count, PAGE_SIZE - count,
				"%d %s %s %s %s %s %s\n", bd->channels[i]->ch_portnum,
				(bd->channels[i]->ch_mostat & UART_MCR_RTS) ? "RTS" : "",
				(bd->channels[i]->ch_mistat & UART_MSR_CTS) ? "CTS" : "",
				(bd->channels[i]->ch_mostat & UART_MCR_DTR) ? "DTR" : "",
				(bd->channels[i]->ch_mistat & UART_MSR_DSR) ? "DSR" : "",
				(bd->channels[i]->ch_mistat & UART_MSR_DCD) ? "DCD" : "",
				(bd->channels[i]->ch_mistat & UART_MSR_RI)  ? "RI"  : "");
		} else {
			count += snprintf(buf + count, PAGE_SIZE - count,
				"%d\n", bd->channels[i]->ch_portnum);
		}
	}
	return count;
}
static DEVICE_ATTR(ports_msignals, S_IRUSR, dgap_ports_msignals_show, NULL);


static ssize_t dgap_ports_iflag_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		count += snprintf(buf + count, PAGE_SIZE - count, "%d %x\n",
			bd->channels[i]->ch_portnum, bd->channels[i]->ch_c_iflag);
	}
	return count;
}
static DEVICE_ATTR(ports_iflag, S_IRUSR, dgap_ports_iflag_show, NULL);


static ssize_t dgap_ports_cflag_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		count += snprintf(buf + count, PAGE_SIZE - count, "%d %x\n",
			bd->channels[i]->ch_portnum, bd->channels[i]->ch_c_cflag);
	}
	return count;
}
static DEVICE_ATTR(ports_cflag, S_IRUSR, dgap_ports_cflag_show, NULL);


static ssize_t dgap_ports_oflag_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		count += snprintf(buf + count, PAGE_SIZE - count, "%d %x\n",
			bd->channels[i]->ch_portnum, bd->channels[i]->ch_c_oflag);
	}
	return count;
}
static DEVICE_ATTR(ports_oflag, S_IRUSR, dgap_ports_oflag_show, NULL);


static ssize_t dgap_ports_lflag_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		count += snprintf(buf + count, PAGE_SIZE - count, "%d %x\n",
			bd->channels[i]->ch_portnum, bd->channels[i]->ch_c_lflag);
	}
	return count;
}
static DEVICE_ATTR(ports_lflag, S_IRUSR, dgap_ports_lflag_show, NULL);


static ssize_t dgap_ports_digi_flag_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		count += snprintf(buf + count, PAGE_SIZE - count, "%d %x\n",
			bd->channels[i]->ch_portnum, bd->channels[i]->ch_digi.digi_flags);
	}
	return count;
}
static DEVICE_ATTR(ports_digi_flag, S_IRUSR, dgap_ports_digi_flag_show, NULL);


static ssize_t dgap_ports_rxcount_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		count += snprintf(buf + count, PAGE_SIZE - count, "%d %ld\n",
			bd->channels[i]->ch_portnum, bd->channels[i]->ch_rxcount);
	}
	return count;
}
static DEVICE_ATTR(ports_rxcount, S_IRUSR, dgap_ports_rxcount_show, NULL);


static ssize_t dgap_ports_txcount_show(struct device *p, struct device_attribute *attr, char *buf)
{
	struct board_t *bd;
	int count = 0;
	int i = 0;

	DGAP_VERIFY_BOARD(p, bd);

	for (i = 0; i < bd->nasync; i++) {
		count += snprintf(buf + count, PAGE_SIZE - count, "%d %ld\n",
			bd->channels[i]->ch_portnum, bd->channels[i]->ch_txcount);
	}
	return count;
}
static DEVICE_ATTR(ports_txcount, S_IRUSR, dgap_ports_txcount_show, NULL);


/* this function creates the sys files that will export each signal status
 * to sysfs each value will be put in a separate filename
 */
void dgap_create_ports_sysfiles(struct board_t *bd)
{
	dev_set_drvdata(&bd->pdev->dev, bd);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_state);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_baud);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_msignals);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_iflag);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_cflag);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_oflag);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_lflag);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_digi_flag);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_rxcount);
	device_create_file(&(bd->pdev->dev), &dev_attr_ports_txcount);
}


/* removes all the sys files created for that port */
void dgap_remove_ports_sysfiles(struct board_t *bd)
{
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_state);
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_baud);
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_msignals);
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_iflag);
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_cflag);
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_oflag);
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_lflag);
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_digi_flag);
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_rxcount);
	device_remove_file(&(bd->pdev->dev), &dev_attr_ports_txcount);
}


int dgap_tty_class_init(void)
{
	dgap_tty_class = class_create(THIS_MODULE, "dgap_tty");
	if (IS_ERR(dgap_tty_class))
		return PTR_ERR(dgap_tty_class);
	return 0;

}


int dgap_tty_class_destroy(void)
{
	if (IS_ERR(dgap_tty_class))
		return PTR_ERR(dgap_tty_class);

	class_destroy(dgap_tty_class);
	return 0;
}


#define DGAP_VERIFY_CHANNEL(p, bd, ch, un)		\
	if (!p)						\
		return (0);				\
	un = (struct un_t *) class_get_devdata(p);	\
	if (!un || un->magic != DGAP_UNIT_MAGIC)	\
		return (0);				\
	ch = un->un_ch;					\
	if (!ch || ch->magic != DGAP_CHANNEL_MAGIC)	\
		return (0);				\
							\
	bd = ch->ch_bd;					\
	if (!bd || bd->magic != DGAP_BOARD_MAGIC)	\
		return (0);				\
	if (bd->state != BOARD_READY)			\
		return (0);				\


static ssize_t dgap_tty_state_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	return snprintf(buf, PAGE_SIZE, "%s", un->un_open_count ? "Open" : "Closed");
}
static CLASS_DEVICE_ATTR(state, S_IRUGO, dgap_tty_state_show, NULL);


static ssize_t dgap_tty_baud_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	return snprintf(buf, PAGE_SIZE, "%d\n", ch->ch_baud_info);
}
static CLASS_DEVICE_ATTR(baud, S_IRUGO, dgap_tty_baud_show, NULL);


static ssize_t dgap_tty_msignals_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	if (ch->ch_open_count) {
		return snprintf(buf, PAGE_SIZE, "%s %s %s %s %s %s\n",
			(ch->ch_mostat & UART_MCR_RTS) ? "RTS" : "",
			(ch->ch_mistat & UART_MSR_CTS) ? "CTS" : "",
			(ch->ch_mostat & UART_MCR_DTR) ? "DTR" : "",
			(ch->ch_mistat & UART_MSR_DSR) ? "DSR" : "",
			(ch->ch_mistat & UART_MSR_DCD) ? "DCD" : "",
			(ch->ch_mistat & UART_MSR_RI)  ? "RI"  : "");
	}
	return 0;
}
static CLASS_DEVICE_ATTR(msignals, S_IRUGO, dgap_tty_msignals_show, NULL);


static ssize_t dgap_tty_iflag_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	return snprintf(buf, PAGE_SIZE, "%x\n", ch->ch_c_iflag);
}
static CLASS_DEVICE_ATTR(iflag, S_IRUGO, dgap_tty_iflag_show, NULL);


static ssize_t dgap_tty_cflag_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	return snprintf(buf, PAGE_SIZE, "%x\n", ch->ch_c_cflag);
}
static CLASS_DEVICE_ATTR(cflag, S_IRUGO, dgap_tty_cflag_show, NULL);


static ssize_t dgap_tty_oflag_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	return snprintf(buf, PAGE_SIZE, "%x\n", ch->ch_c_oflag);
}
static CLASS_DEVICE_ATTR(oflag, S_IRUGO, dgap_tty_oflag_show, NULL);


static ssize_t dgap_tty_lflag_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	return snprintf(buf, PAGE_SIZE, "%x\n", ch->ch_c_lflag);
}
static CLASS_DEVICE_ATTR(lflag, S_IRUGO, dgap_tty_lflag_show, NULL);


static ssize_t dgap_tty_digi_flag_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	return snprintf(buf, PAGE_SIZE, "%x\n", ch->ch_digi.digi_flags);
}
static CLASS_DEVICE_ATTR(digi_flag, S_IRUGO, dgap_tty_digi_flag_show, NULL);


static ssize_t dgap_tty_rxcount_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	return snprintf(buf, PAGE_SIZE, "%ld\n", ch->ch_rxcount);
}
static CLASS_DEVICE_ATTR(rxcount, S_IRUGO, dgap_tty_rxcount_show, NULL);


static ssize_t dgap_tty_txcount_show(struct class_device *class_dev, char *buf)
{
	struct board_t *bd;
	struct channel_t *ch;
	struct un_t *un;

	DGAP_VERIFY_CHANNEL(class_dev, bd, ch, un);

	return snprintf(buf, PAGE_SIZE, "%ld\n", ch->ch_txcount);
}
static CLASS_DEVICE_ATTR(txcount, S_IRUGO, dgap_tty_txcount_show, NULL);


void dgap_tty_register_device(struct un_t *un, struct device *device)
{
	struct class_device *dgap_class_member;
	dev_t dev;

	if (un->un_type == DGAP_PRINT) {
		dev = MKDEV(un->un_ch->ch_bd->PrintDriver.major,
			    un->un_ch->ch_bd->PrintDriver.minor_start +
			    un->un_ch->ch_portnum);

		dgap_class_member = class_device_create(dgap_tty_class, NULL, dev, device,
			"prn%d%c", un->un_ch->ch_bd->boardnum + 1, 
			'a' + un->un_ch->ch_portnum);
	}
	else {
		dev = MKDEV(un->un_ch->ch_bd->SerialDriver.major,
			    un->un_ch->ch_bd->SerialDriver.minor_start + 
			    un->un_ch->ch_portnum);

		dgap_class_member = class_device_create(dgap_tty_class, NULL, dev, device,
			"ttyn%d%c", un->un_ch->ch_bd->boardnum + 1, 
			'a' + un->un_ch->ch_portnum);
	}

	if (IS_ERR(dgap_class_member))
		return;

	class_set_devdata(dgap_class_member, un);
	class_device_create_file(dgap_class_member, &class_device_attr_state);
	class_device_create_file(dgap_class_member, &class_device_attr_baud);
	class_device_create_file(dgap_class_member, &class_device_attr_msignals);
	class_device_create_file(dgap_class_member, &class_device_attr_iflag);
	class_device_create_file(dgap_class_member, &class_device_attr_cflag);
	class_device_create_file(dgap_class_member, &class_device_attr_oflag);
	class_device_create_file(dgap_class_member, &class_device_attr_lflag);
	class_device_create_file(dgap_class_member, &class_device_attr_digi_flag);
	class_device_create_file(dgap_class_member, &class_device_attr_rxcount);
	class_device_create_file(dgap_class_member, &class_device_attr_txcount);
}


void dgap_tty_unregister_device(struct un_t *un)
{
	if (un->un_type == DGAP_PRINT) {
		class_device_destroy(dgap_tty_class, MKDEV(un->un_ch->ch_bd->PrintDriver.major,
			un->un_ch->ch_bd->PrintDriver.minor_start + un->un_ch->ch_portnum));
	}
	else {
		class_device_destroy(dgap_tty_class, MKDEV(un->un_ch->ch_bd->SerialDriver.major,
			un->un_ch->ch_bd->SerialDriver.minor_start + un->un_ch->ch_portnum));
	}
}


#endif  /* DGNC_SYSFS */
